"""
HTTP request related code.
"""
import base64
import datetime
import json
import logging
import os
import shlex
import subprocess
from typing import Optional

try:
    import google.auth
    from google.auth.transport.requests import Request as GoogleAuthRequest

    google_auth_installed = True
except ImportError:
    google_auth_installed = False
try:
    from requests_oauthlib import OAuth2Session

    oidc_auth_installed = True
except ImportError:
    oidc_auth_installed = False

import requests.adapters

from http import HTTPStatus
from urllib.parse import urlparse

from .exceptions import HTTPError, PyKubeError
from .utils import jsonpath_installed, jsonpath_parse, join_url_path
from .config import KubeConfig

from . import __version__

DEFAULT_HTTP_TIMEOUT = 10  # seconds
EXPIRY_SKEW_PREVENTION_DELAY = datetime.timedelta(minutes=5)
UTC = datetime.timezone.utc
LOG = logging.getLogger(__name__)


class KubernetesHTTPAdapter(requests.adapters.HTTPAdapter):

    # _do_send: the actual send method of HTTPAdapter
    # it can be overwritten in unit tests to mock the actual HTTP calls
    _do_send = requests.adapters.HTTPAdapter.send

    def __init__(self, kube_config: KubeConfig, **kwargs):
        self.kube_config = kube_config

        super().__init__(**kwargs)

    def _persist_credentials(self, config, opts):
        user_name = config.contexts[config.current_context]["user"]
        user = [u["user"] for u in config.doc["users"] if u["name"] == user_name][0]
        auth_config = user["auth-provider"].setdefault("config", {})
        auth_config.update(opts)
        config.persist_doc()
        config.reload()

    def _auth_gcp(self, request, token, expiry, config):
        original_request = request.copy()

        credentials = google.auth.default(
            scopes=[
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/userinfo.email",
            ]
        )[0]
        credentials.token = token
        credentials.expiry = expiry

        should_persist = not credentials.valid

        auth_request = GoogleAuthRequest()
        credentials.before_request(
            auth_request, request.method, request.url, request.headers
        )

        if should_persist and config:
            auth_opts = {
                "access-token": credentials.token,
                "expiry": credentials.expiry,
            }
            self._persist_credentials(config, auth_opts)

        def retry(send_kwargs):
            credentials.refresh(auth_request)
            response = self.send(original_request, **send_kwargs)
            if response.ok and config:
                auth_opts = {
                    "access-token": credentials.token,
                    "expiry": credentials.expiry,
                }
                self._persist_credentials(config, auth_opts)
            return response

        return retry

    def _is_valid_jwt(self, token):
        """Validate JWT token for correctness and near expiration"""
        if not token:
            return False
        reserved_characters = frozenset(["=", "+", "/"])
        if any(char in token for char in reserved_characters):
            # Invalid jwt, as it contains url-unsafe chars
            return False
        parts = token.split(".")
        if len(parts) != 3:  # Not a valid JWT
            return False
        padding = (4 - len(parts[1]) % 4) * "="
        if len(padding) == 3:
            # According to spec, 3 padding characters cannot occur
            # in a valid jwt
            # https://tools.ietf.org/html/rfc7515#appendix-C
            return False
        jwt_attributes = json.loads(
            base64.b64decode(parts[1] + padding).decode("utf-8")
        )
        expire = jwt_attributes.get("exp")
        # allow missing exp, but deny tokens that are about to expire soon
        return expire is None or (
            datetime.datetime.fromtimestamp(expire, tz=UTC)
            - EXPIRY_SKEW_PREVENTION_DELAY
        ) > datetime.datetime.utcnow().replace(tzinfo=UTC)

    def _refresh_oidc_token(self, config):
        if not oidc_auth_installed:
            raise ImportError(
                "missing dependencies for OIDC token refresh support "
                "(try pip install pykube-ng[oidc]"
            )
        auth_config = config.user["auth-provider"]["config"]
        if "idp-certificate-authority" in auth_config:
            verify = auth_config["idp-certificate-authority"].filename()
        else:
            verify = None
        oauth = OAuth2Session()
        discovery = oauth.get(
            f"{auth_config['idp-issuer-url']}/.well-known/openid-configuration",
            verify=verify,
            timeout=DEFAULT_HTTP_TIMEOUT,
            withhold_token=True,
        )

        if discovery.status_code != HTTPStatus.OK:
            raise PyKubeError(
                f"Failed to discover OpenID token endpoint - "
                f"HTTP {discovery.status_code}: {discovery.text}"
            )
        discovery = discovery.json()
        refresh = oauth.refresh_token(
            token_url=discovery["token_endpoint"],
            refresh_token=auth_config["refresh-token"],
            client_id=auth_config["client-id"],
            client_secret=auth_config.get("client-secret"),
            verify=verify,
            timeout=DEFAULT_HTTP_TIMEOUT,
        )
        auth_opts = {
            "id-token": refresh["id_token"],
            "refresh-token": refresh["refresh_token"],
        }
        self._persist_credentials(config, auth_opts)

    def send(self, request, **kwargs):
        if "kube_config" in kwargs:
            config = kwargs.pop("kube_config")
        else:
            config = self.kube_config

        _retry_attempt = kwargs.pop("_retry_attempt", 0)
        retry_func = self._setup_request_auth(config, request, kwargs)
        self._setup_request_certificates(config, request, kwargs)

        response = self._do_send(request, **kwargs)

        _retry_status_codes = {HTTPStatus.UNAUTHORIZED}

        if (
            response.status_code in _retry_status_codes
            and retry_func
            and _retry_attempt < 2
        ):
            send_kwargs = {"_retry_attempt": _retry_attempt + 1, "kube_config": config}
            send_kwargs.update(kwargs)
            return retry_func(send_kwargs=send_kwargs)

        return response

    def _setup_request_auth(self, config, request, kwargs):
        """
        Set up authorization for the request.

        Return an optional function to use as a retry manager if the initial request fails
        with an unauthorized error.
        """
        if "Authorization" in request.headers:
            # request already has some auth header (e.g. Bearer token)
            # don't modify/overwrite it
            return None

        if config.user.get("token"):
            request.headers["Authorization"] = "Bearer {}".format(config.user["token"])
            return None

        if "exec" in config.user:
            exec_conf = config.user["exec"]

            api_version = exec_conf["apiVersion"]
            if api_version == "client.authentication.k8s.io/v1alpha1":
                cmd_env_vars = dict(os.environ)
                for env_var in exec_conf.get("env") or []:
                    cmd_env_vars[env_var["name"]] = env_var["value"]

                output = subprocess.check_output(
                    [exec_conf["command"]] + exec_conf["args"], env=cmd_env_vars
                )

                parsed_out = json.loads(output)
                token = parsed_out["status"]["token"]
            else:
                raise NotImplementedError(
                    f"auth exec api version {api_version} not implemented"
                )

            request.headers["Authorization"] = "Bearer {}".format(token)
            return None

        if config.user.get("username") and config.user.get("password"):
            request.prepare_auth((config.user["username"], config.user["password"]))
            return None

        if "auth-provider" in config.user:
            auth_provider = config.user["auth-provider"]
            if auth_provider.get("name") == "gcp":
                dependencies = [google_auth_installed, jsonpath_installed]
                if not all(dependencies):
                    raise ImportError(
                        "missing dependencies for GCP support (try pip install pykube-ng[gcp]"
                    )
                auth_config = auth_provider.get("config", {})
                if "cmd-path" in auth_config:
                    output = subprocess.check_output(
                        [auth_config["cmd-path"]] + shlex.split(auth_config["cmd-args"])
                    )
                    parsed = json.loads(output)
                    token = jsonpath_parse(auth_config["token-key"], parsed)
                    expiry = datetime.datetime.strptime(
                        jsonpath_parse(auth_config["expiry-key"], parsed),
                        "%Y-%m-%dT%H:%M:%SZ",
                    )
                    retry_func = self._auth_gcp(request, token, expiry, None)
                else:
                    retry_func = self._auth_gcp(
                        request,
                        auth_config.get("access-token"),
                        auth_config.get("expiry"),
                        config,
                    )
                return retry_func
            elif auth_provider.get("name") == "oidc":
                auth_config = auth_provider.get("config", {})
                if not self._is_valid_jwt(auth_config.get("id-token")):
                    try:
                        self._refresh_oidc_token(config)
                    # ignoring all exceptions, rely on retries
                    except Exception as oidc_exc:
                        LOG.warning(f"Failed to refresh OpenID token: {oidc_exc}")

                # not using auth_config handle here as the config might have
                # been reloaded during token refresh
                request.headers["Authorization"] = "Bearer {}".format(
                    config.user["auth-provider"]["config"]["id-token"]
                )

        return None

    def _setup_request_certificates(self, config, request, kwargs):
        if "client-certificate" in config.user:
            kwargs["cert"] = (
                config.user["client-certificate"].filename(),
                config.user["client-key"].filename(),
            )
        # setup certificate verification
        if os.environ.get("PYKUBE_SSL_CERTIFICATE_AUTHORITY") is not None:
            kwargs["verify"] = os.environ.get("PYKUBE_SSL_CERTIFICATE_AUTHORITY")
        elif "certificate-authority" in config.cluster:
            kwargs["verify"] = config.cluster["certificate-authority"].filename()
        elif "insecure-skip-tls-verify" in config.cluster:
            kwargs["verify"] = not config.cluster["insecure-skip-tls-verify"]


class HTTPClient:
    """
    Client for interfacing with the Kubernetes API.
    """

    http_adapter_cls = KubernetesHTTPAdapter

    def __init__(
        self,
        config: KubeConfig,
        timeout: float = DEFAULT_HTTP_TIMEOUT,
        dry_run: bool = False,
        verify: bool = True,
        http_adapter: Optional[requests.adapters.HTTPAdapter] = None,
    ):
        """
        Creates a new instance of the HTTPClient.

        :Parameters:
           - `config`: The configuration instance
        """
        self.config = config
        self.timeout = timeout
        self.url = self.config.cluster["server"]
        self.dry_run = dry_run

        session = requests.Session()
        session.headers["User-Agent"] = f"pykube-ng/{__version__}"
        if not http_adapter:
            http_adapter = self.http_adapter_cls(self.config)
        session.mount("https://", http_adapter)
        session.mount("http://", http_adapter)
        self.session = session
        self.session.verify = verify

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        pr = urlparse(value)
        self._url = pr.geturl()

    @property
    def version(self):
        """
        Get Kubernetes API version
        """
        response = self.get(version="", base="/version")
        response.raise_for_status()
        data = response.json()
        return (data["major"], data["minor"])

    def resource_list(self, api_version):
        cached_attr = f"_cached_resource_list_{api_version}"
        if not hasattr(self, cached_attr):
            r = self.get(version=api_version)
            r.raise_for_status()
            setattr(self, cached_attr, r.json())
        return getattr(self, cached_attr)

    def get_kwargs(self, **kwargs) -> dict:
        """
        Creates a full URL to request based on arguments.

        :Parametes:
           - `kwargs`: All keyword arguments to build a kubernetes API endpoint
        """
        version = kwargs.pop("version", "v1")
        if version == "v1":
            base = kwargs.pop("base", "/api")
        elif "/" in version:
            base = kwargs.pop("base", "/apis")
        else:
            if "base" not in kwargs:
                raise TypeError("unknown API version; base kwarg must be specified.")
            base = kwargs.pop("base")
        if version.startswith("/"):
            # for compatibility with pykube-ng 20.1.0 when calling api.get(version="/apis"):
            # posixpath.join() was throwing away everything before the first "absolute" path (i.e. starting with a slash)
            bits = [version]
        else:
            bits = [base, version]
        # Overwrite (default) namespace from context if it was set
        if "namespace" in kwargs:
            n = kwargs.pop("namespace")
            if n is not None:
                if n:
                    namespace = n
                else:
                    namespace = self.config.namespace
                if namespace:
                    bits.extend(["namespaces", namespace])
        url = kwargs.get("url", "")
        bits.append(url)
        kwargs["url"] = self.url + join_url_path(*bits, join_empty=True)
        if "timeout" not in kwargs:
            # apply default HTTP timeout
            kwargs["timeout"] = self.timeout
        if self.dry_run:
            # Add http query param for dryRun
            params = kwargs.get("params", {})
            params["dryRun"] = "All"
            kwargs["params"] = params
        return kwargs

    def raise_for_status(self, resp):
        try:
            resp.raise_for_status()
        except Exception:
            # attempt to provide a more specific exception based around what
            # Kubernetes returned as the error.
            if resp.headers["content-type"] == "application/json":
                payload = resp.json()
                if payload["kind"] == "Status":
                    raise HTTPError(resp.status_code, payload["message"])
            raise

    def request(self, *args, **kwargs):
        """
        Makes an API request based on arguments.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.request(*args, **self.get_kwargs(**kwargs))

    def get(self, *args, **kwargs):
        """
        Executes an HTTP GET.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.get(*args, **self.get_kwargs(**kwargs))

    def options(self, *args, **kwargs):
        """
        Executes an HTTP OPTIONS.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.options(*args, **self.get_kwargs(**kwargs))

    def head(self, *args, **kwargs):
        """
        Executes an HTTP HEAD.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.head(*args, **self.get_kwargs(**kwargs))

    def post(self, *args, **kwargs):
        """
        Executes an HTTP POST.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.post(*args, **self.get_kwargs(**kwargs))

    def put(self, *args, **kwargs):
        """
        Executes an HTTP PUT.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.put(*args, **self.get_kwargs(**kwargs))

    def patch(self, *args, **kwargs):
        """
        Executes an HTTP PATCH.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.patch(*args, **self.get_kwargs(**kwargs))

    def delete(self, *args, **kwargs):
        """
        Executes an HTTP DELETE.

        :Parameters:
           - `args`: Non-keyword arguments
           - `kwargs`: Keyword arguments
        """
        return self.session.delete(*args, **self.get_kwargs(**kwargs))



import json
from collections import namedtuple
from typing import Union
from urllib.parse import urlencode

from .exceptions import ObjectDoesNotExist
from .http import HTTPClient


all_ = object()
everything = object()
now = object()


class Table:
    """
    Tabular resource representation
    See https://kubernetes.io/docs/reference/using-api/api-concepts/#receiving-resources-as-tables
    """

    def __init__(self, api_obj_class, obj: dict):
        assert obj["kind"] == "Table"
        self.api_obj_class = api_obj_class
        self.obj = obj

    def __repr__(self) -> str:
        return "<Table of {kind} at {address}>".format(
            kind=self.api_obj_class.kind, address=hex(id(self))
        )

    @property
    def columns(self):
        return self.obj["columnDefinitions"]

    @property
    def rows(self):
        return self.obj["rows"]


class BaseQuery:
    def __init__(self, api: HTTPClient, api_obj_class, namespace: str = None):
        self.api = api
        self.api_obj_class = api_obj_class
        self.namespace = namespace
        self.selector = everything
        self.field_selector = everything

    def __repr__(self) -> str:
        return "<Query of {kind} at {address}>".format(
            kind=self.api_obj_class.kind, address=hex(id(self))
        )

    def all(self) -> "BaseQuery":
        return self._clone()

    def filter(
        self,
        namespace: str = None,
        selector: Union[str, dict] = None,
        field_selector: Union[str, dict] = None,
    ) -> "BaseQuery":
        """
        Filter objects by namespace, labels, or fields

        :param namespace: Namespace to filter by (pass pykube.all to get objects in all namespaces)
        :param selector: Label selector, can be a dictionary of label names/values
        """
        clone = self._clone()
        if namespace is not None:
            clone.namespace = namespace
        if selector is not None:
            clone.selector = selector
        if field_selector is not None:
            clone.field_selector = field_selector
        return clone

    def _clone(self, cls=None):
        if cls is None:
            cls = self.__class__
        clone = cls(self.api, self.api_obj_class, namespace=self.namespace)
        clone.selector = self.selector
        clone.field_selector = self.field_selector
        return clone

    def _build_api_url(self, params: dict = None):
        if params is None:
            params = {}
        if self.selector is not everything:
            params["labelSelector"] = as_selector(self.selector)  # type: ignore
        if self.field_selector is not everything:
            params["fieldSelector"] = as_selector(self.field_selector)  # type: ignore
        query_string = urlencode(params)
        return "{}{}".format(
            self.api_obj_class.endpoint,
            f"?{query_string}" if query_string else "",
        )


class Query(BaseQuery):
    def get_by_name(self, name: str):
        """
        Get object by name, raises ObjectDoesNotExist if not found
        """
        kwargs = {
            "url": f"{self.api_obj_class.endpoint}/{name}",
            "namespace": self.namespace,
        }
        if self.api_obj_class.base:
            kwargs["base"] = self.api_obj_class.base
        if self.api_obj_class.version:
            kwargs["version"] = self.api_obj_class.version
        r = self.api.get(**kwargs)
        if not r.ok:
            if r.status_code == 404:
                raise ObjectDoesNotExist(f"{name} does not exist.")
            self.api.raise_for_status(r)
        return self.api_obj_class(self.api, r.json())

    def get(self, *args, **kwargs):
        """
        Get a single object by name, namespace, label, ..
        """
        if "name" in kwargs:
            return self.get_by_name(kwargs["name"])
        clone = self.filter(*args, **kwargs)
        num = len(clone)
        if num == 1:
            return clone.query_cache["objects"][0]
        if not num:
            raise ObjectDoesNotExist("get() returned zero objects")
        raise ValueError("get() more than one object; use filter")

    def get_or_none(self, *args, **kwargs):
        """
        Get object by name, return None if not found
        """
        try:
            return self.get(*args, **kwargs)
        except ObjectDoesNotExist:
            return None

    def watch(self, since=None, *, params=None):
        query = self._clone(WatchQuery)
        query.params = params
        if since is now:
            query.resource_version = self.response["metadata"]["resourceVersion"]
        elif since is not None:
            query.resource_version = since
        return query

    def execute(self, **kwargs):
        kwargs["url"] = self._build_api_url()
        if self.api_obj_class.base:
            kwargs["base"] = self.api_obj_class.base
        if self.api_obj_class.version:
            kwargs["version"] = self.api_obj_class.version
        if self.namespace is not None and self.namespace is not all_:
            kwargs["namespace"] = self.namespace
        r = self.api.get(**kwargs)
        r.raise_for_status()
        return r

    def as_table(self) -> Table:
        """
        Execute query and return result as Table (similar to what kubectl does)
        See https://kubernetes.io/docs/reference/using-api/api-concepts/#receiving-resources-as-tables
        """
        response = self.execute(
            headers={"Accept": "application/json;as=Table;v=v1beta1;g=meta.k8s.io"}
        )
        return Table(self.api_obj_class, response.json())

    def iterator(self):
        """
        Execute the API request and return an iterator over the objects. This
        method does not use the query cache.
        """
        for obj in self.execute().json().get("items") or []:
            yield self.api_obj_class(self.api, obj)

    @property
    def query_cache(self):
        if not hasattr(self, "_query_cache"):
            cache = {"objects": []}
            cache["response"] = self.execute().json()
            for obj in cache["response"].get("items") or []:
                cache["objects"].append(self.api_obj_class(self.api, obj))
            self._query_cache = cache
        return self._query_cache

    def __len__(self):
        return len(self.query_cache["objects"])

    def __iter__(self):
        return iter(self.query_cache["objects"])

    @property
    def response(self):
        return self.query_cache["response"]


class WatchQuery(BaseQuery):
    def __init__(self, *args, **kwargs):
        self.resource_version = kwargs.pop("resource_version", None)
        self.params = None
        super(WatchQuery, self).__init__(*args, **kwargs)
        self._response = None

    def object_stream(self):
        params = dict(self.params or {})  # shallow clone for local use
        params["watch"] = "true"
        if self.resource_version is not None:
            params["resourceVersion"] = self.resource_version
        kwargs = {"url": self._build_api_url(params=params), "stream": True}
        if self.namespace is not all_:
            kwargs["namespace"] = self.namespace
        if self.api_obj_class.version:
            kwargs["version"] = self.api_obj_class.version
        r = self.api.get(**kwargs)
        self.api.raise_for_status(r)
        self._response = r
        WatchEvent = namedtuple("WatchEvent", "type object")
        for line in r.iter_lines():
            we = json.loads(line.decode("utf-8"))
            yield WatchEvent(
                type=we["type"], object=self.api_obj_class(self.api, we["object"])
            )

    def __iter__(self):
        return iter(self.object_stream())

    @property
    def response(self):
        return self._response


def as_selector(value: Union[str, dict]) -> str:
    if isinstance(value, str):
        return value
    s = []
    for k, v in value.items():
        bits = k.split("__")
        assert len(bits) <= 2, "too many __ in selector"
        if len(bits) == 1:
            label = bits[0]
            op = "eq"
        else:
            label = bits[0]
            op = bits[1]
        # map operator to selector
        if op == "eq":
            s.append(f"{label}={v}")
        elif op == "neq":
            s.append(f"{label} != {v}")
        elif op == "in":
            s.append("{} in ({})".format(label, ",".join(sorted(v))))
        elif op == "notin":
            s.append("{} notin ({})".format(label, ",".join(sorted(v))))
        else:
            raise ValueError(f"{op} is not a valid comparison operator")
    return ",".join(s)



"""
Configuration code.
"""
import base64
import copy
import hashlib
import os
import tempfile
from pathlib import Path
from typing import Optional

import yaml

from pykube import exceptions


def _join_host_port(host, port):
    """Adapted golang's net.JoinHostPort"""
    template = "{}:{}"
    host_requires_bracketing = ":" in host or "%" in host
    if host_requires_bracketing:
        template = "[{}]:{}"
    return template.format(host, port)


class KubeConfig:
    """
    Main configuration class.
    """

    filepath = None

    @classmethod
    def from_service_account(
        cls, path="/var/run/secrets/kubernetes.io/serviceaccount", **kwargs
    ):
        """
        Construct KubeConfig from in-cluster service account.
        """
        service_account_dir = Path(path)

        with service_account_dir.joinpath("token").open() as fp:
            token = fp.read()

        with service_account_dir.joinpath("namespace").open() as fp:
            namespace = fp.read()

        host = os.environ.get("PYKUBE_KUBERNETES_SERVICE_HOST")
        if host is None:
            host = os.environ["KUBERNETES_SERVICE_HOST"]
        port = os.environ.get("PYKUBE_KUBERNETES_SERVICE_PORT")
        if port is None:
            port = os.environ["KUBERNETES_SERVICE_PORT"]
        doc = {
            "clusters": [
                {
                    "name": "self",
                    "cluster": {
                        "server": "https://" + _join_host_port(host, port),
                        "certificate-authority": str(
                            service_account_dir.joinpath("ca.crt")
                        ),
                    },
                }
            ],
            "users": [{"name": "self", "user": {"token": token}}],
            "contexts": [
                {
                    "name": "self",
                    "context": {
                        "cluster": "self",
                        "user": "self",
                        "namespace": namespace,
                    },
                }
            ],
            "current-context": "self",
        }
        self = cls(doc, **kwargs)
        return self

    @classmethod
    def from_file(cls, filename=None, **kwargs):
        """
        Creates an instance of the KubeConfig class from a kubeconfig file.

        :param filename: The full path to the configuration file. Defaults to ~/.kube/config
        """
        if not filename:
            filename = os.getenv("KUBECONFIG", "~/.kube/config")
        filepath = Path(filename).expanduser()
        if not filepath.is_file():
            raise exceptions.PyKubeError(
                "Configuration file {} not found".format(filename)
            )
        with filepath.open() as f:
            doc = yaml.safe_load(f.read())
        self = cls(doc, **kwargs)
        self.filepath = filepath
        return self

    @classmethod
    def from_env(cls):
        """
        Convenience function to create an instance of KubeConfig from the current environment.

        First tries to use in-cluster ServiceAccount, then tries default ~/.kube/config (or KUBECONFIG)
        """
        try:
            # running in cluster
            config = cls.from_service_account()
        except FileNotFoundError:
            # not running in cluster => load local ~/.kube/config
            config = cls.from_file()
        return config

    @classmethod
    def from_url(cls, url, **kwargs):
        """
        Creates an instance of the KubeConfig class from a single URL (useful
        for interacting with kubectl proxy).
        """
        doc = {
            "clusters": [{"name": "self", "cluster": {"server": url}}],
            "contexts": [{"name": "self", "context": {"cluster": "self"}}],
            "current-context": "self",
        }
        self = cls(doc, **kwargs)
        return self

    def __init__(self, doc, current_context=None):
        """
        Creates an instance of the KubeConfig class.
        """
        self.doc = doc
        self._current_context = None
        if current_context is not None:
            self.set_current_context(current_context)
        elif "current-context" in doc and doc["current-context"]:
            self.set_current_context(doc["current-context"])

    def set_current_context(self, value):
        """
        Sets the context to the provided value.

        :Parameters:
           - `value`: The value for the current context
        """
        self._current_context = value

    @property
    def kubeconfig_path(self) -> Optional[Path]:
        """
        Returns the path to kubeconfig file, if it exists
        """
        if not hasattr(self, "filepath"):
            return None
        return self.filepath

    @property
    def kubeconfig_file(self) -> Optional[str]:
        """
        Returns the path to kubeconfig file as string, if it exists
        """
        if not hasattr(self, "filepath"):
            return None
        return str(self.filepath)

    @property
    def current_context(self):
        if self._current_context is None:
            raise exceptions.PyKubeError(
                "current context not set; call set_current_context"
            )
        return self._current_context

    @property
    def clusters(self):
        """
        Returns known clusters by exposing as a read-only property.
        """
        if not hasattr(self, "_clusters"):
            cs = {}
            for cr in self.doc["clusters"]:
                cs[cr["name"]] = c = copy.deepcopy(cr["cluster"])
                if "server" not in c:
                    c["server"] = "http://localhost"
                BytesOrFile.maybe_set(c, "certificate-authority", self.kubeconfig_path)
            self._clusters = cs
        return self._clusters

    @property
    def users(self):
        """
        Returns known users by exposing as a read-only property.
        """
        if not hasattr(self, "_users"):
            us = {}
            if "users" in self.doc:
                for ur in self.doc["users"]:
                    us[ur["name"]] = u = copy.deepcopy(ur["user"])
                    BytesOrFile.maybe_set(u, "client-certificate", self.kubeconfig_path)
                    BytesOrFile.maybe_set(u, "client-key", self.kubeconfig_path)
                    if "auth-provider" in u:
                        BytesOrFile.maybe_set(
                            u["auth-provider"]["config"],
                            "idp-certificate-authority",
                            self.kubeconfig_path,
                        )
            self._users = us
        return self._users

    @property
    def contexts(self):
        """
        Returns known contexts by exposing as a read-only property.
        """
        if not hasattr(self, "_contexts"):
            cs = {}
            for cr in self.doc["contexts"]:
                cs[cr["name"]] = copy.deepcopy(cr["context"])
            self._contexts = cs
        return self._contexts

    @property
    def cluster(self):
        """
        Returns the current selected cluster by exposing as a
        read-only property.
        """
        return self.clusters[self.contexts[self.current_context]["cluster"]]

    @property
    def user(self):
        """
        Returns the current user set by current context
        """
        return self.users.get(self.contexts[self.current_context].get("user", ""), {})

    @property
    def namespace(self) -> str:
        """
        Returns the current context namespace by exposing as a read-only property.
        """
        return self.contexts[self.current_context].get("namespace", "default")

    def persist_doc(self):

        if not self.kubeconfig_path:
            # Config was provided as string, not way to persit it
            return
        with self.kubeconfig_path.open("w") as f:
            yaml.safe_dump(
                self.doc,
                f,
                encoding="utf-8",
                allow_unicode=True,
                default_flow_style=False,
            )

    def reload(self):
        if hasattr(self, "_users"):
            delattr(self, "_users")
        if hasattr(self, "_contexts"):
            delattr(self, "_contexts")
        if hasattr(self, "_clusters"):
            delattr(self, "_clusters")


class BytesOrFile:
    """
    Implements the same interface for files and byte input.
    """

    @classmethod
    def maybe_set(cls, d, key, kubeconfig_path):
        file_key = key
        data_key = "{}-data".format(key)
        if data_key in d:
            d[file_key] = cls(data=d[data_key], kubeconfig_path=kubeconfig_path)
            del d[data_key]
        elif file_key in d:
            d[file_key] = cls(filename=d[file_key], kubeconfig_path=kubeconfig_path)

    def __init__(self, filename=None, data=None, kubeconfig_path=None):
        """
        Creates a new instance of BytesOrFile.

        :Parameters:
           - `filename`: A full path to a file
           - `data`: base64 encoded bytes
        """
        self._path = None
        self._bytes = None
        if filename is not None and data is not None:
            raise TypeError("filename or data kwarg must be specified, not both")
        elif filename is not None:

            path = Path(filename)
            # If relative path is given, should be made absolute with respect to the directory of the kube config
            # https://kubernetes.io/docs/concepts/configuration/organize-cluster-access-kubeconfig/#file-references
            if not path.is_absolute():
                if kubeconfig_path:
                    path = kubeconfig_path.parent.joinpath(path)
                else:
                    raise exceptions.PyKubeError(
                        "{} passed as relative path, but cannot determine location of kube config".format(
                            filename
                        )
                    )

            if not path.is_file():
                raise exceptions.PyKubeError(
                    "'{}' file does not exist".format(filename)
                )
            self._path = path
        elif data is not None:
            self._bytes = base64.b64decode(data)
        else:
            raise TypeError("filename or data kwarg must be specified")

    def bytes(self):
        """
        Returns the provided data as bytes.
        """
        if self._path:
            with self._path.open("rb") as f:
                return f.read()
        else:
            return self._bytes

    def filename(self):
        """
        Returns the provided data as a file location.
        """
        if not self._path:

            m = hashlib.md5()
            m.update(self._bytes)

            path = Path(f"{tempfile.gettempdir()}/pykube-ng.{m.hexdigest()}.crt")
            if not path.exists() or path.stat().st_size == 0:
                with open(path, "wb") as f:
                    f.write(self._bytes)

            self._path = path

        return str(self._path)



import copy
import json
from inspect import getmro
from typing import Any
from typing import Optional
from typing import Type
from urllib.parse import urlencode

from requests import Response

from .exceptions import ObjectDoesNotExist
from .http import HTTPClient
from .mixins import ReplicatedMixin
from .mixins import ScalableMixin
from .query import Query
from .utils import join_url_path
from .utils import obj_merge


class ObjectManager:
    def __call__(self, api: HTTPClient, namespace: str = None):
        if namespace is None and NamespacedAPIObject in getmro(self.api_obj_class):
            namespace = api.config.namespace
        return Query(api, self.api_obj_class, namespace=namespace)

    def __get__(self, obj, api_obj_class: Type):
        assert obj is None, "cannot invoke objects on resource object."
        self.api_obj_class = api_obj_class
        return self


class APIObject:
    """
    Baseclass for all Kubernetes API objects
    """

    objects = ObjectManager()
    base = None

    def __init__(self, api: HTTPClient, obj: dict):
        self.api = api
        self.set_obj(obj)

    def set_obj(self, obj: dict):
        self.obj = obj
        self._original_obj = copy.deepcopy(obj)

    def __repr__(self):
        return f"<{self.kind} {self.name}>"

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        """
        Name of the Kubernetes resource (metadata.name)

        Name must be unique within a namespace. Is required when creating resources, although some resources may allow a client to request the generation
        of an appropriate name automatically. Name is primarily intended for creation idempotence and configuration definition.
        Cannot be updated. More info: http://kubernetes.io/docs/user-guide/identifiers#names
        """
        return self.obj["metadata"]["name"]

    @property
    def namespace(self) -> Optional[str]:
        return None

    @property
    def metadata(self):
        return self.obj["metadata"]

    @property
    def labels(self) -> dict:
        """
        Labels of the Kubernetes resource (metadata.labels)

        Map of string keys and values that can be used to organize and categorize (scope and select) objects.
        May match selectors of replication controllers and services. More info: http://kubernetes.io/docs/user-guide/labels
        """
        return self.obj["metadata"].setdefault("labels", {})

    @property
    def annotations(self) -> dict:
        """
        Annotations of the Kubernetes resource (metadata.annotations)

        Annotations is an unstructured key value map stored with a resource that may be set by external tools to store and retrieve arbitrary metadata.
        They are not queryable and should be preserved when modifying objects. More info: http://kubernetes.io/docs/user-guide/annotations
        """
        return self.obj["metadata"].setdefault("annotations", {})

    def api_kwargs(self, **kwargs):
        kw = {}
        # Construct url for api request
        obj_list = kwargs.pop("obj_list", False)
        if obj_list:
            kw["url"] = self.endpoint
        else:
            subresource = kwargs.pop("subresource", None) or ""
            operation = kwargs.pop("operation", "")
            kw["url"] = join_url_path(self.endpoint, self.name, subresource, operation)

        params = kwargs.pop("params", None)
        if params is not None:
            query_string = urlencode(params)
            kw["url"] = "{}{}".format(
                kw["url"], f"?{query_string}" if query_string else ""
            )
        if self.base:
            kw["base"] = self.base
        kw["version"] = self.version
        if self.namespace is not None:
            kw["namespace"] = self.namespace
        kw.update(kwargs)
        return kw

    def exists(self, ensure=False):
        r = self.api.get(**self.api_kwargs())
        if r.status_code not in {200, 404}:
            self.api.raise_for_status(r)
        if not r.ok:
            if ensure:
                raise ObjectDoesNotExist(f"{self.name} does not exist.")
            else:
                return False
        return True

    def create(self):
        r = self.api.post(**self.api_kwargs(data=json.dumps(self.obj), obj_list=True))
        self.api.raise_for_status(r)
        self.set_obj(r.json())

    def reload(self):
        r = self.api.get(**self.api_kwargs())
        self.api.raise_for_status(r)
        self.set_obj(r.json())

    def watch(self):
        return (
            self.__class__.objects(self.api, namespace=self.namespace)
            .filter(field_selector={"metadata.name": self.name})
            .watch()
        )

    def patch(self, strategic_merge_patch, *, subresource=None):
        """
        Patch the Kubernetes resource by calling the API with a "strategic merge" patch.
        """
        r = self.api.patch(
            **self.api_kwargs(
                subresource=subresource,
                headers={"Content-Type": "application/merge-patch+json"},
                data=json.dumps(strategic_merge_patch),
            )
        )
        self.api.raise_for_status(r)
        self.set_obj(r.json())

    def update(self, is_strategic=True, *, subresource=None):
        """
        Update the Kubernetes resource by calling the API (patch)
        """
        self.obj = obj_merge(self.obj, self._original_obj, is_strategic)
        self.patch(self.obj, subresource=subresource)

    def delete(self, propagation_policy: str = None):
        """
        Delete the Kubernetes resource by calling the API.

        The parameter propagation_policy defines whether to cascade the delete. It can be "Foreground", "Background" or "Orphan".
        See https://kubernetes.io/docs/concepts/workloads/controllers/garbage-collection/#setting-the-cascading-deletion-policy
        """
        if propagation_policy:
            options = {"propagationPolicy": propagation_policy}
        else:
            options = {}
        r = self.api.delete(**self.api_kwargs(data=json.dumps(options)))
        if r.status_code != 404:
            self.api.raise_for_status(r)


class NamespacedAPIObject(APIObject):
    @property
    def namespace(self) -> str:
        """
        Namespace scope of the Kubernetes resource (metadata.namespace)

        Namespace defines the space within each name must be unique.
        Cannot be updated. More info: http://kubernetes.io/docs/user-guide/namespaces
        """
        if self.obj["metadata"].get("namespace"):
            return self.obj["metadata"]["namespace"]
        else:
            return self.api.config.namespace


def object_factory(api, api_version, kind) -> Type[APIObject]:
    """
    Dynamically builds a Python class for the given Kubernetes object in an API.

    For example:

        api = pykube.HTTPClient(...)
        NetworkPolicy = pykube.object_factory(api, "networking.k8s.io/v1", "NetworkPolicy")

    This enables construction of any Kubernetes object kind without explicit support
    from pykube.

    Currently, the HTTPClient passed to this function will not be bound to the returned type.
    It is planned to fix this, but in the mean time pass it as you would normally.
    """
    resource_list = api.resource_list(api_version)
    try:
        resource = next(
            resource
            for resource in resource_list["resources"]
            if resource["kind"] == kind
        )
    except StopIteration:
        raise ValueError("unknown resource kind {!r}".format(kind)) from None
    base = NamespacedAPIObject if resource["namespaced"] else APIObject
    return type(
        kind,
        (base,),
        {"version": api_version, "endpoint": resource["name"], "kind": kind},
    )


class ConfigMap(NamespacedAPIObject):
    version = "v1"
    endpoint = "configmaps"
    kind = "ConfigMap"


class CronJob(NamespacedAPIObject):
    version = "batch/v1beta1"
    endpoint = "cronjobs"
    kind = "CronJob"


class DaemonSet(NamespacedAPIObject):
    version = "apps/v1"
    endpoint = "daemonsets"
    kind = "DaemonSet"


class Deployment(NamespacedAPIObject, ReplicatedMixin, ScalableMixin):
    version = "apps/v1"
    endpoint = "deployments"
    kind = "Deployment"

    @property
    def ready(self):
        return (
            self.obj["status"]["observedGeneration"]
            >= self.obj["metadata"]["generation"]
            and self.obj["status"]["updatedReplicas"] == self.replicas
        )

    def rollout_undo(self, target_revision=None):
        """Produces same action as kubectl rollout undo deployment command.
        Input variable is revision to rollback to (in kubectl, --to-revision)
        """
        if target_revision is None:
            revision = {}
        else:
            revision = {"revision": target_revision}

        params = {
            "kind": "DeploymentRollback",
            "apiVersion": self.version,
            "name": self.name,
            "rollbackTo": revision,
        }

        kwargs = {
            "version": self.version,
            "namespace": self.namespace,
            "operation": "rollback",
        }
        r = self.api.post(**self.api_kwargs(data=json.dumps(params), **kwargs))
        r.raise_for_status()
        return r.text


class Endpoint(NamespacedAPIObject):
    version = "v1"
    endpoint = "endpoints"
    kind = "Endpoint"


class Event(NamespacedAPIObject):
    version = "v1"
    endpoint = "events"
    kind = "Event"


class LimitRange(NamespacedAPIObject):
    version = "v1"
    endpoint = "limitranges"
    kind = "LimitRange"


class ResourceQuota(NamespacedAPIObject):
    version = "v1"
    endpoint = "resourcequotas"
    kind = "ResourceQuota"


class ServiceAccount(NamespacedAPIObject):
    version = "v1"
    endpoint = "serviceaccounts"
    kind = "ServiceAccount"


class Ingress(NamespacedAPIObject):
    version = "networking.k8s.io/v1"
    endpoint = "ingresses"
    kind = "Ingress"


class Job(NamespacedAPIObject, ScalableMixin):
    version = "batch/v1"
    endpoint = "jobs"
    kind = "Job"
    scalable_attr = "parallelism"

    @property
    def parallelism(self):
        return self.obj["spec"]["parallelism"]

    @parallelism.setter
    def parallelism(self, value):
        self.obj["spec"]["parallelism"] = value


class Namespace(APIObject):
    version = "v1"
    endpoint = "namespaces"
    kind = "Namespace"


class Node(APIObject):
    version = "v1"
    endpoint = "nodes"
    kind = "Node"

    @property
    def unschedulable(self):
        if "unschedulable" in self.obj["spec"]:
            return self.obj["spec"]["unschedulable"]
        return False

    @unschedulable.setter
    def unschedulable(self, value):
        self.obj["spec"]["unschedulable"] = value
        self.update()

    def cordon(self):
        self.unschedulable = True

    def uncordon(self):
        self.unschedulable = False


class Pod(NamespacedAPIObject):
    version = "v1"
    endpoint = "pods"
    kind = "Pod"

    @property
    def ready(self):
        cs = self.obj["status"].get("conditions", [])
        condition = next((c for c in cs if c["type"] == "Ready"), None)
        return condition is not None and condition["status"] == "True"

    def logs(
        self,
        container=None,
        pretty=None,
        previous=False,
        since_seconds=None,
        since_time=None,
        timestamps=False,
        tail_lines=None,
        limit_bytes=None,
    ):
        """
        Produces the same result as calling kubectl logs pod/<pod-name>.
        Check parameters meaning at
        http://kubernetes.io/docs/api-reference/v1/operations/,
        part 'read log of the specified Pod'. The result is plain text.
        """
        log_call = "log"
        params = {}
        if container is not None:
            params["container"] = container
        if pretty is not None:
            params["pretty"] = pretty
        if previous:
            params["previous"] = "true"
        if since_seconds is not None and since_time is None:
            params["sinceSeconds"] = int(since_seconds)
        elif since_time is not None and since_seconds is None:
            params["sinceTime"] = since_time
        if timestamps:
            params["timestamps"] = "true"
        if tail_lines is not None:
            params["tailLines"] = int(tail_lines)
        if limit_bytes is not None:
            params["limitBytes"] = int(limit_bytes)

        query_string = urlencode(params)
        log_call += f"?{query_string}" if query_string else ""
        kwargs = {
            "version": self.version,
            "namespace": self.namespace,
            "operation": log_call,
        }
        r = self.api.get(**self.api_kwargs(**kwargs))
        r.raise_for_status()
        return r.text


class ReplicationController(NamespacedAPIObject, ReplicatedMixin, ScalableMixin):
    version = "v1"
    endpoint = "replicationcontrollers"
    kind = "ReplicationController"

    @property
    def ready(self):
        return (
            self.obj["status"]["observedGeneration"]
            >= self.obj["metadata"]["generation"]
            and self.obj["status"]["readyReplicas"] == self.replicas
        )


class ReplicaSet(NamespacedAPIObject, ReplicatedMixin, ScalableMixin):
    version = "apps/v1"
    endpoint = "replicasets"
    kind = "ReplicaSet"


class Secret(NamespacedAPIObject):
    version = "v1"
    endpoint = "secrets"
    kind = "Secret"


class Service(NamespacedAPIObject):
    version = "v1"
    endpoint = "services"
    kind = "Service"

    def proxy_http_request(
        self, method: str, path: str, port: Optional[int] = None, **kwargs: Any
    ) -> Response:
        """Issue a HTTP request with specific HTTP method to proxy of a Service.
        Args:
            :param method: The http request method e.g. 'GET', 'POST' etc.
            :param path: The URI path for the request.
            :param port: This value can be used to override the
            default (first defined) port used to connect to the Service.
            :param kwargs: Keyword arguments for the proxy_http_get function.
            They are the same as for requests.models.Request object.
        Returns:
            The requests.Response object.
        """
        if port is None:
            port = self.obj["spec"]["ports"][0]["port"]
        kwargs["url"] = f"services/{self.name}:{port}/proxy/{path}"
        kwargs["namespace"] = self.namespace
        kwargs["version"] = self.version
        return self.api.request(method, **kwargs)

    def proxy_http_get(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        """Issue a HTTP GET request to proxy of a Service.
        Args:
            :param path: The URI path for the request.
            :param port: This value can be used to override the
            default (first defined) port used to connect to the Service.
            :param kwargs: Keyword arguments for the proxy_http_get function.
            They are the same as for requests.models.Request object
            plus the additional 'port' kwarg, which can be used to override the
            default (first defined) port used to connect to the Service.
        Returns:
            The requests.Response object.
        """
        return self.proxy_http_request("GET", path, port, **kwargs)

    def proxy_http_post(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        """Issue a HTTP POST request to proxy of a Service.
        Args:
            :param path: The URI path for the request.
            :param port: This value can be used to override the
            default (first defined) port used to connect to the Service.
            :param kwargs: Keyword arguments for the proxy_http_get function.
            They are the same as for requests.models.Request object
            plus the additional 'port' kwarg, which can be used to override the
            default (first defined) port used to connect to the Service.
        Returns:
            The requests.Response object.
        """
        return self.proxy_http_request("POST", path, port, **kwargs)

    def proxy_http_put(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        """Issue a HTTP PUT request to proxy of a Service.
        Args:
            :param path: The URI path for the request.
            :param port: This value can be used to override the
            default (first defined) port used to connect to the Service.
            :param kwargs: Keyword arguments for the proxy_http_get function.
            They are the same as for requests.models.Request object
            plus the additional 'port' kwarg, which can be used to override the
            default (first defined) port used to connect to the Service.
        Returns:
            The requests.Response object.
        """
        return self.proxy_http_request("PUT", path, port, **kwargs)

    def proxy_http_delete(
        self, path: str, port: Optional[int] = None, **kwargs
    ) -> Response:
        """Issue a HTTP DELETE request to proxy of a Service.
        Args:
            :param path: The URI path for the request.
            :param port: This value can be used to override the
            default (first defined) port used to connect to the Service.
            :param kwargs: Keyword arguments for the proxy_http_get function.
            They are the same as for requests.models.Request object
            plus the additional 'port' kwarg, which can be used to override the
            default (first defined) port used to connect to the Service.
        Returns:
            The requests.Response object.
        """
        return self.proxy_http_request("DELETE", path, port, **kwargs)


class PersistentVolume(APIObject):
    version = "v1"
    endpoint = "persistentvolumes"
    kind = "PersistentVolume"


class PersistentVolumeClaim(NamespacedAPIObject):
    version = "v1"
    endpoint = "persistentvolumeclaims"
    kind = "PersistentVolumeClaim"


class HorizontalPodAutoscaler(NamespacedAPIObject):
    version = "autoscaling/v1"
    endpoint = "horizontalpodautoscalers"
    kind = "HorizontalPodAutoscaler"


class StatefulSet(NamespacedAPIObject, ReplicatedMixin, ScalableMixin):
    version = "apps/v1"
    endpoint = "statefulsets"
    kind = "StatefulSet"


class Role(NamespacedAPIObject):
    version = "rbac.authorization.k8s.io/v1"
    endpoint = "roles"
    kind = "Role"


class RoleBinding(NamespacedAPIObject):
    version = "rbac.authorization.k8s.io/v1"
    endpoint = "rolebindings"
    kind = "RoleBinding"


class ClusterRole(APIObject):
    version = "rbac.authorization.k8s.io/v1"
    endpoint = "clusterroles"
    kind = "ClusterRole"


class ClusterRoleBinding(APIObject):
    version = "rbac.authorization.k8s.io/v1"
    endpoint = "clusterrolebindings"
    kind = "ClusterRoleBinding"


class PodSecurityPolicy(APIObject):
    version = "policy/v1beta1"
    endpoint = "podsecuritypolicies"
    kind = "PodSecurityPolicy"


class PodDisruptionBudget(NamespacedAPIObject):
    version = "policy/v1beta1"
    endpoint = "poddisruptionbudgets"
    kind = "PodDisruptionBudget"


class CustomResourceDefinition(APIObject):
    version = "apiextensions.k8s.io/v1"
    endpoint = "customresourcedefinitions"
    kind = "CustomResourceDefinition"



import argparse
import code

import pykube

# import readline to support line editing within console session.
try:
    import readline  # noqa: F401
except ImportError:
    pass


def main(argv=None):
    """
    Run the interactive Pykube console (usually invoked via python3 -m pykube)
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--kubeconfig", help="Path to the kubeconfig file to use", metavar="PATH"
    )
    parser.add_argument(
        "--context", help="The name of the kubeconfig context to used", metavar="NAME"
    )
    parser.add_argument(
        "-c", help="Python program passed in as string", metavar="SCRIPT"
    )
    args = parser.parse_args(argv)

    config = pykube.KubeConfig.from_file(args.kubeconfig)

    if args.context:
        config.set_current_context(args.context)

    api = pykube.HTTPClient(config)

    context = {
        "__name__": "__console__",
        "pykube": pykube,
        "config": config,
        "api": api,
    }
    for k, v in vars(pykube).items():
        if k[0] != "_" and k[0] == k[0].upper():
            context[k] = v

    banner = f"""Pykube v{pykube.__version__}, loaded "{config.filepath}" with context "{config.current_context}".

    Example commands:
      [d.name for d in Deployment.objects(api)]              # get names of deployments in default namespace
      list(DaemonSet.objects(api, namespace='kube-system'))  # list daemonsets in "kube-system"
      Pod.objects(api).get_by_name('mypod').labels           # labels of pod "mypod"

    Use Ctrl-D to exit"""

    console = code.InteractiveConsole(locals=context)
    if args.c:
        console.runsource(args.c)
    else:
        console.interact(banner)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pykube']

package_data = \
{'': ['*']}

install_requires = \
['pyyaml', 'requests>=2.12']

extras_require = \
{'gcp': ['google-auth', 'jsonpath-ng'],
 'oidc': ['requests-oauthlib>=1.3.0,<2.0.0']}

setup_kwargs = {
    'name': 'pykube-ng',
    'version': '21.10.0',
    'description': 'Python client library for Kubernetes',
    'long_description': 'pykube-ng\n=========\n\n.. image:: https://img.shields.io/travis/hjacobs/pykube.svg\n   :target: https://travis-ci.org/hjacobs/pykube\n   :alt: Build status\n\n.. image:: https://coveralls.io/repos/github/hjacobs/pykube/badge.svg?branch=master;_=1\n   :target: https://coveralls.io/github/hjacobs/pykube?branch=master\n   :alt: Code Coverage\n\n.. image:: https://readthedocs.org/projects/pykube/badge/?version=latest\n   :target: https://pykube.readthedocs.io/\n   :alt: Documentation\n\n.. image:: https://img.shields.io/pypi/v/pykube-ng.svg\n   :target: https://pypi.python.org/pypi/pykube-ng/\n   :alt: PyPI version\n\n.. image:: https://img.shields.io/pypi/pyversions/pykube-ng.svg\n   :target: https://pypi.python.org/pypi/pykube-ng/\n   :alt: Python versions\n\n.. image:: https://img.shields.io/badge/license-apache-blue.svg\n   :target: https://pypi.python.org/pypi/pykube-ng/\n   :alt: Apache License\n\n.. image:: https://img.shields.io/badge/calver-YY.MM.MICRO-22bfda.svg\n   :target: http://calver.org/\n   :alt: CalVer\n\nPykube (pykube-ng) is a lightweight Python 3.6+ client library for Kubernetes.\n\nThis is a fork of `kelproject/pykube <https://github.com/kelproject/pykube>`_ which is no longer maintained (archived). Here the original text of the pykube README:\n\n    Kel is an open source Platform as a Service (PaaS) from Eldarion, Inc. that\n    makes it easy to manage web application deployment and hosting through the\n    entire lifecycle from development through testing to production. It adds\n    components and tools on top of Kubernetes that help developers manage their\n    application infrastructure. Kel builds on Eldarion\'s 7+ years experience running\n    one of the leading Python and Django PaaSes.\n    For more information about Kel, see `kelproject.com`_ or follow us on Twitter\n    `@projectkel`_.\n\n.. _kelproject.com: http://kelproject.com/\n.. _@projectkel: https://twitter.com/projectkel\n\nFeatures\n--------\n\n* HTTP interface using requests using kubeconfig for authentication\n* Python native querying of Kubernetes API objects\n\nInstallation\n------------\n\nTo install pykube, use pip::\n\n    pip install pykube-ng\n\n\nInteractive Console\n-------------------\n\nThe ``pykube`` library module can be run as an interactive console locally for quick exploration.\nIt will automatically load ``~/.kube/config`` to provide the ``api`` object, and it loads pykube classes (``Deployment``, ``Pod``, ..) into local context:\n\n.. code-block:: bash\n\n    python3 -m pykube\n    >>> [d.name for d in Deployment.objects(api)]\n\n\nUsage\n-----\n\nQuery for all ready pods in a custom namespace:\n\n.. code:: python\n\n    import operator\n    import pykube\n\n    api = pykube.HTTPClient(pykube.KubeConfig.from_file())\n    pods = pykube.Pod.objects(api).filter(namespace="gondor-system")\n    ready_pods = filter(operator.attrgetter("ready"), pods)\n\nAccess any attribute of the Kubernetes object:\n\n.. code:: python\n\n    pod = pykube.Pod.objects(api).filter(namespace="gondor-system").get(name="my-pod")\n    pod.obj["spec"]["containers"][0]["image"]\n\nSelector query:\n\n.. code:: python\n\n    pods = pykube.Pod.objects(api).filter(\n        namespace="gondor-system",\n        selector={"gondor.io/name__in": {"api-web", "api-worker"}},\n    )\n    pending_pods = pykube.objects.Pod.objects(api).filter(\n        field_selector={"status.phase": "Pending"}\n    )\n\nWatch query:\n\n.. code:: python\n\n    watch = pykube.Job.objects(api, namespace="gondor-system")\n    watch = watch.filter(field_selector={"metadata.name": "my-job"}).watch()\n\n    # watch is a generator:\n    for watch_event in watch:\n        print(watch_event.type) # \'ADDED\', \'DELETED\', \'MODIFIED\'\n        print(watch_event.object) # pykube.Job object\n\nCreate a Deployment:\n\n.. code:: python\n\n    obj = {\n        "apiVersion": "apps/v1",\n        "kind": "Deployment",\n        "metadata": {\n            "name": "my-deploy",\n            "namespace": "gondor-system"\n        },\n        "spec": {\n            "replicas": 3,\n            "selector": {\n                "matchLabels": {\n                    "app": "nginx"\n                }\n            },\n            "template": {\n                "metadata": {\n                    "labels": {\n                        "app": "nginx"\n                    }\n                },\n                "spec": {\n                    "containers": [\n                        {\n                            "name": "nginx",\n                            "image": "nginx",\n                            "ports": [\n                                {"containerPort": 80}\n                            ]\n                        }\n                    ]\n                }\n            }\n        }\n    }\n    pykube.Deployment(api, obj).create()\n\nDelete a Deployment:\n\n.. code:: python\n\n    obj = {\n        "apiVersion": "apps/v1",\n        "kind": "Deployment",\n        "metadata": {\n            "name": "my-deploy",\n            "namespace": "gondor-system"\n        }\n    }\n    pykube.Deployment(api, obj).delete()\n\nCheck server version:\n\n.. code:: python\n\n    api = pykube.HTTPClient(pykube.KubeConfig.from_file())\n    api.version\n\n\nRequirements\n------------\n\n* Python 3.6+\n* requests (included in ``install_requires``)\n* PyYAML (included in ``install_requires``)\n\n\nLocal Development\n-----------------\n\nYou can run pykube against your current kubeconfig context, e.g. local Minikube_:\n\n.. code-block:: bash\n\n    poetry install\n    poetry run python3\n    >>> import pykube\n    >>> config = pykube.KubeConfig.from_file()\n    >>> api = pykube.HTTPClient(config)\n    >>> list(pykube.Deployment.objects(api))\n\nTo run PEP8 (flake8) checks and unit tests including coverage report:\n\n.. code-block:: bash\n\n    make test\n\n\nLicense\n-------\n\nThe code in this project is licensed under the Apache License, version 2.0\n(included in this repository under LICENSE).\n\n\nContributing\n------------\n\nEasiest way to contribute is to provide feedback! We would love to hear what you like and what you think is missing.\nCreate an issue or `ping try_except_ on Twitter`_.\n\nPRs are welcome. Please also have a look at `issues labeled with "help wanted"`_.\n\n\nCode of Conduct\n----------------\n\nIn order to foster a kind, inclusive, and harassment-free community, this project follows the `Contributor Covenant Code of Conduct`_.\n\n.. _Contributor Covenant Code of Conduct: http://contributor-covenant.org/version/1/4/\n\n\n.. _ping try_except_ on Twitter: https://twitter.com/try_except_\n.. _issues labeled with "help wanted": https://codeberg.org/hjacobs/pykube-ng/issues?q=is%3Aissue+is%3Aopen+label%3A%22help+wanted%22\n.. _Minikube: https://github.com/kubernetes/minikube\n',
    'author': 'Eldarion, Inc.',
    'author_email': 'development@eldarion.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://codeberg.org/hjacobs/pykube-ng',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import re
from typing import List

try:
    from jsonpath_ng import parse as jsonpath

    jsonpath_installed = True
except ImportError:
    jsonpath_installed = False

from itertools import zip_longest


empty = object()


def obj_merge(obj, original_obj, is_strategic=True):
    c = {}
    for k, v in obj.items():
        if k not in original_obj:
            c[k] = v
        else:
            c[k] = obj_check(v, original_obj[k], is_strategic)

    if is_strategic is True:
        for k, v in original_obj.items():
            if k not in obj:
                c[k] = v
    return c


def obj_check(obj_value, original_obj_value, is_strategic=True):
    check_result = None
    if not isinstance(obj_value, type(original_obj_value)):
        check_result = obj_value
    else:
        if isinstance(obj_value, dict):
            check_result = obj_merge(obj_value, original_obj_value, is_strategic)

        elif isinstance(obj_value, list):
            if is_strategic:
                res_list = []
                for x, y in zip_longest(obj_value, original_obj_value, fillvalue=empty):
                    if x is empty:
                        res_list.append(y)
                    elif y is empty:
                        res_list.append(x)
                    else:
                        res_list.append(obj_check(x, y, is_strategic))
                check_result = res_list
            else:
                check_result = obj_value
        else:
            check_result = obj_value
    return check_result


def jsonpath_parse(template, obj):
    def repl(m):
        path = m.group(2)
        if not path.startswith("$"):
            path = "$" + path
        return jsonpath(path).find(obj)[0].value

    return re.sub(r"(\{([^\}]*)\})", repl, template)


def join_url_path(*components, join_empty: bool = False) -> str:
    """Join given URL path components and return absolute path starting with '/'."""
    new_comps: List[str] = []
    for comp in components:
        comp = comp.strip("/")
        if comp in ("", "."):
            continue
        else:
            new_comps.append(comp)

    if components and components[-1] == "" and join_empty:
        trailing_slash = True
    elif components and components[-1] == "/":
        trailing_slash = True
    else:
        trailing_slash = False

    if trailing_slash:
        new_comps.append("")

    path = "/".join(new_comps)
    return "/" + path
